from revu.queues.github import GithubQueue
from revu.queue import Review

import tmuxp
import readline
import cmd


class Session:
    def __init__(self, projects):
        self.projects = projects
        self.project_map = {x.name: x for x in self.projects}

    def get(self, project):
        return self.project_map[project]

    def pulls(self, project):
        if isinstance(project, str):
            project = self.project_map[project]

        gq = GithubQueue()
        for changeset in gq.pulls([project.repo]):
            yield Review(project.path, changeset)


class TmuxController:
    def __init__(self, session_name):
        self.session_name = session_name
        self.server = tmuxp.Server()
        try:
            self.session = self.server.new_session(
                session_name=self.session_name,
                kill_session=False,
                detatch=True,
            )
        except tmuxp.exc.TmuxSessionExists:
            self.session = self.server.findWhere({
                "session_name": session_name
            })

    def _clean(self):
        for window in self.session.list_windows():
            window.kill_window()

    def window(self, window_name, start_directory, *args, **kwargs):
        # self._clean()  # XXX: Implement this correctly.
        return self.session.new_window(
            window_name=window_name,
            start_directory=start_directory,
            attach=True)



class SessionREPL(cmd.Cmd):
    def __init__(self, projects):
        self.projects = projects
        self.session = Session(self.projects)
        self.tmux_controller = TmuxController("revu")

        self.review = None
        self.project = None
        self.prs = []

        print("Please run: `tmux attach-session -t revu`")
        print("")
        print("")

        super(SessionREPL, self).__init__()

    @property
    def prompt(self):
        ps1 = ""

        if self.project:
            ps1 += "({}) ".format(self.project.repo)

        if self.review:
            ps1 += "#{} ".format(self.review.pr.number)

        ps1 += "λ "
        return ps1

    def do_workon(self, line):
        self.project = self.session.get(line)

    do_w = do_workon

    def do_prs(self, line):
        if self.project is None:
            print("Please select a project with `workon` first.")
            return
        self.prs = list(self.session.pulls(self.project))

        print("Open Pull Requests")
        print("==================")
        for i, pr in enumerate(self.prs):
            print("{}) - GH #{pr.number} - {pr.title}".format(i, pr=pr.pr))

    def do_review(self, line):
        self.review = self.prs[int(line)]
        mc = self.review.checkout()
        if mc:
            print("Euch, there was a merge conflict.")
        print("Ready for human intervention. - check the tmux pane".format(
            self.review.repo
        ))
        self.tmux_controller.window(
            "pr-{}".format(self.review.pr.number),
            self.review.repo
        )

    do_r = do_review

    def do_comment(self, line):
        if self.review is None:
            print("Need to be on a review")
            return
        self.review.comment(line)

    def do_merge(self, line):
        if self.review is None:
            print("Need to be on a review")
            return
        self.review.merge()

    def do_push(self, line):
        if self.review is None:
            print("Need to be on a review")
            return
        self.review.push()

    def do_list(self, line):
        print("Known projects")
        print("==============")
        for project in self.projects:
            print("{project.name}".format(project=project))
        print("")

    do_l = do_list

    def do_EOF(self, line):
        return True



from sh import git
import sh

import os
import shutil
import tempfile
import contextlib
import configparser


class Queue:
    pass


class Project:
    def __init__(self, name, repo, path):
        self.name = name
        self.repo = repo
        self.path = path

    @classmethod
    def from_config(cls, path):
        config = configparser.ConfigParser()
        config.read([os.path.expanduser(path)])
        for section, name in ((config[x], x) for x in config.sections()):
            yield Project(name=name, **section)


class PullRequest:  # Implied this is a Git repo
    def __init__(self, number, title, body, origin, oref, remote, rref):
        self.number = number
        self.title = title
        self.body = body
        self.origin = origin
        self.oref = oref
        self.remote = remote
        self.rref = rref

    def __str__(self):
        return "<PullRequest: {remote}/{rref} -> {origin}/{oref}".format(
            **self.__dict__
        )


class Review:
    def __init__(self, repo, pullrequest):
        self.pr = pullrequest
        self.repo = repo
        self._git = None

    def log(self, message):
        print(message)

    def checkout(self):
        self.log("Cloning origin; please wait")

        self._git = git.bake(C=self.repo)
        self._git.checkout(self.pr.oref)

        self.log("Pulling down PR {number}".format(number=self.pr.number))
        # This is where we get GitHub specific; please fix this
        # for future review plugins.
        self._git.fetch(
            "origin",
            "refs/pull/{number}/head:pr/{number}".format(number=self.pr.number))

        self.log("Checking out")
        self._git.checkout("pr/{number}".format(number=self.pr.number))
        self.log("Merging root ref in")
        conflict = False

        try:
            self._git.merge(self.pr.oref)
        except sh.ErrorReturnCode as e:
            self.log("Merge conflict! Yikes!")
            self.log(e.stdout.decode('utf-8'))
            conflict = True

        return conflict

    def merge(self):
        self.log("Checking out origin ref")
        self._git.checkout(self.pr.oref)
        self.log("Merging in the PR branch")
        self._git.merge("pr/{number}".format(number=self.pr.number))
        self.log("Done.")

    def push(self):
        self._git.push("origin", self.pr.oref)

    def comment(self, message):
        issue = self.pr._repo.issue(self.pr._pr.number)
        comment = issue.create_comment(message)



from revu import __appname__, __version__
from setuptools import setup


long_description = ""

setup(
    name=__appname__,
    version=__version__,
    packages=[
        'revu',
        'revu.queues',
    ],
    author="Paul Tagliamonte",
    author_email="tag@pault.ag",
    long_description=long_description,
    description='n/a',
    license="Expat",
    url="http://pault.ag/",
    platforms=['any'],
    entry_points={
        'console_scripts': [
            'revu = revu.cli:main',
        ]
    }
)



import os
import github3
from ..queue import Queue, PullRequest


class GithubQueue(Queue):
    def __init__(self):
        self._github = github3.GitHub("paultag", self.get_key())  # Erm.

    def get_key(self):
        key = os.environ.get("GITHUB_API_KEY", None)
        if key:
            return key

        try:
            with open(os.path.expanduser("~/.github.key"), 'r') as fd:
                key = fd.read().strip()
        except IOError:
            raise ValueError("No key defined")

        return key

    def _get_pr(self, user, repo):
        repo = self._github.repository(user, repo)
        requests = repo.iter_pulls()
        for request in requests:
            yield (request, repo)

    def _get_prs(self, repos):
        for repo in repos:
            yield from self._get_pr(*repo.split("/", 1))

    def pulls(self, repos):
        for pr, repo in self._get_prs(repos):

            remote = "git://github.com/{}/{}.git".format(*pr.head.repo)
            origin = "git://github.com/{}/{}.git".format(*pr.base.repo)

            npr = PullRequest(
                pr.number, pr.title, pr.body,
                origin,
                pr.base.ref,
                remote,
                pr.head.ref)

            npr._pr = pr
            npr._repo = repo  # XXX: Subclass
            yield npr
